(* 
 * Copyright 2012 Damien Guichard  
 *  
 * Licensed under the EUPL, Version 1.1 only (the "Licence");
 * You may not use this work except in compliance with the Licence. 
 * You may obtain a copy of the Licence at: 
 *  
 * http://ec.europa.eu/idabc/eupl5  
 *  
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the Licence is distributed on an "AS IS" basis, 
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. 
 * See the Licence for the specific language governing permissions
 * and limitations under the Licence. 
 *) 

open Lex

module Loop(ER:Types.ER_DiGraph with type name=Datas.TaxonomyTree.t)
:
sig 
end
=
struct
	let print_concept =
		let module Taxo = Datas.TaxonomyTree in function
		| ER.Entity(t) ->
			print_string (Taxo.to_string t) 
		| ER.EntityN(t,n) ->
			print_string (Taxo.to_string t); print_char ' '; print_string n
		| ER.EntityS(t,s) ->
			print_string (Taxo.to_string t); print_char ' '; 
			print_char '"'; print_string s; print_char '"'
		| ER.EntityI(t,i) -> 
			print_string (Taxo.to_string t); print_char ' '; print_int i
		| ER.EntityR(t,i) -> 
			print_string (Taxo.to_string t);
			if i >= 0 then (print_string " #"; print_int i)
		| ER.EntityF(t,x) -> 
			print_string (Taxo.to_string t); print_char ' '; print_float x
	let print_vertex c mark = function
		| ER.Vertex ->
			print_char '[';
			print_concept c;
			if mark <> "" then (print_char '*'; print_string mark);
			print_char ']'
		| ER.Source e -> print_string e.ER.src_mark
		| ER.Destination e -> print_string e.ER.dst_mark
	let rec print_pattern =
		let module Taxo = Datas.TaxonomyTree in function
		| [] ->
			print_char '.'; print_newline ()
		| ({ER.src=src;rel;dst;src_mark;dst_mark},sv,dv)::tl ->
			print_char '(';
			print_vertex src src_mark sv;
			print_char ' '; print_string (Taxo.to_string rel); print_char ' ';
			print_vertex dst dst_mark dv;
			print_char ')';
			print_pattern tl		
	let print_pattern p () =
		print_pattern p
	let main =
		let module Dict = Datas.Dictionary in
		let module Taxo = Datas.TaxonomyTree in
		let ent,top = Taxo.singleton "Entity" in
		let rel,_ = Taxo.singleton "Relation" in
		object
			val mutable ent = ent
			val mutable rel = rel
			val mutable kdb = ER.empty
			val mutable dic = Dict.empty
			val mutable env = Dict.empty
			val mutable entity = ER.Entity top  		
			val mutable select = false  		
			val src_mark = "" val dst_mark = "" val no_mark = ""
		initializer
			print_endline "Entity-Relationship interactive calculator."; 
			print_endline "version 0.2c by Damien Guichard."; 
			while true do
				print_string  "# "; flush stdout;
				let lex = new Lex.make stdin in
				let demand_entity () =
					let t = Taxo.from_string lex#demand_name ent in
					if not (lex#promised_mono ']') then
					ignore(lex#granted_mono ':');
					match () with
					| _ when lex#promised_name ->
						ER.EntityN(t,lex#demand_name)
					| _ when lex#granted_mono '+' ->
						if lex#promised_int then ER.EntityI(t,lex#demand_int)
						else ER.EntityF(t,lex#demand_float)
					| _ when lex#granted_mono '-' ->
						if lex#promised_int then ER.EntityI(t,-lex#demand_int)
						else ER.EntityF(t,-.lex#demand_float)
					| _ when lex#promised_int ->
						ER.EntityI(t,lex#demand_int)
					| _ when lex#promised_float ->
						ER.EntityF(t,lex#demand_float)
					| _ when lex#granted_mono '#' ->
						ER.EntityR(t,lex#demand_int)
					| _ when lex#promised_mono '"' ->
						ER.EntityS(t,lex#demand_string)
					| _ ->
						if select then ER.Entity t
						else ER.EntityR(t,Taxo.unique t)
				in						
				let demand_join_entity () =
					match () with 
					| _ when lex#granted_mono '[' ->
						let c = demand_entity () in 
						if lex#granted_mono '*' then
						dic <- Dict.add lex#demand_name c dic;
						lex#demand_mono ']'; c					
					| _ ->
						Dict.find lex#demand_name dic
				in
				let rec demand_join () =
					let p = lex#granted_mono '(' in
					let src = demand_join_entity ()
					and rel = Taxo.from_string lex#demand_name rel
					and dst = demand_join_entity () in 
					if p then lex#demand_mono ')';
					ER.({src;rel;dst;src_mark;dst_mark})::
					(if lex#promised_mono '.' then [] else demand_join ())
				in
				let demand_select_entity () =
					match () with 
					| _ when lex#granted_mono '[' ->
						let c = demand_entity () 
						and mark =
							if lex#granted_mono '*' then lex#demand_name
							else no_mark
						in lex#demand_mono ']'; c,mark,ER.Vertex					
					| _ ->
						entity,no_mark,Dict.find lex#demand_name env
				in
				let rec demand_select () =
					let p = lex#granted_mono '(' in
					let src,src_mark,src_v = demand_select_entity ()
					and rel = Taxo.from_string lex#demand_name rel
					and dst,dst_mark,dst_v = demand_select_entity () in 
					if p then lex#demand_mono ')';
					let edge = ER.({src;rel;dst;src_mark;dst_mark}) in
					if src_mark != no_mark then
					env <- Dict.add src_mark (ER.Source edge) env; 
					if dst_mark != no_mark then
					env <- Dict.add dst_mark (ER.Destination edge) env; 
					(edge,src_v,dst_v)::
					(if lex#promised_mono '.' then [] else demand_select ())
				in
				let rec demand_relations acc =
					if lex#promised_mono '.' then acc
					else if lex#granted_mono '-' then
						let s = lex#demand_name in
						lex#demand_mono '-'; demand_relations (s::acc)
					else 
						let s = lex#demand_name in demand_relations
						begin
							if lex#granted_mono '|' then lex#demand_name::s::acc
							else s::acc
						end
				in
				let rec demand_entities acc =
					if lex#promised_mono '.' then acc
					else demand_entities (lex#demand_name::acc)
				in
				let assert_mono c =
					if not (lex#promised_mono c) then lex#demand_mono c
				in
				try
					dic <- Dict.empty;
					env <- Dict.empty;
					match () with
					| _ when lex#granted_word "ontology" -> 
						lex#demand_mono '[';
						let t,top = Taxo.singleton lex#demand_name in
						lex#demand_mono ']'; lex#demand_mono '(';
						let r,_ = Taxo.singleton lex#demand_name in
						lex#demand_mono ')';
						assert_mono '.';
						ent <- t; rel <- r; kdb <- ER.empty; entity <- ER.Entity top
					| _ when lex#granted_word "super-type" ->
						if lex#granted_mono '(' then begin
							let str = lex#demand_name in
							lex#demand_mono ')';
							let parent = Taxo.from_string str rel in
							Taxo.add_childs_to (demand_relations []) parent rel 
						end else begin
							lex#demand_mono '[';
							let str = lex#demand_name in
							lex#demand_mono ']';
							let parent = Taxo.from_string str ent in
							Taxo.add_childs_to (demand_entities []) parent ent 
						end 
					| _ when lex#granted_word "join" ->
						kdb <- ER.insert (demand_join()) kdb
					| _ when lex#granted_word "select" ->
						select <- true; 
						let pat = demand_select () in
						ER.select (print_pattern pat) pat kdb;
						select <- false 
					| _ when lex#granted_word "load" -> 
						assert_mono '"';
						let str = lex#demand_string in
						assert_mono '.';
						let ch = open_in_bin str in
						let t,r,k = Marshal.from_channel ch in
						ent <- t; rel <- r; kdb <- k; 
						close_in ch
					| _ when lex#granted_word "save" -> 
						assert_mono '"';
						let str = lex#demand_string in
						assert_mono '.';
						let ch = open_out_bin str in
						Marshal.to_channel ch (ent,rel,kdb) ER.marshal_flags;
						close_out ch
					| _ when lex#granted_word "quit" -> 
						assert_mono '.'; exit 0
					| _ -> failwith "command expected"
				with
				| Lex.Demand_denied (p,d) ->
					Lex.print_position p;
					Lex.print_denied d
				| Failure msg | Sys_error msg ->
					Lex.print_position lex#demand_position;
					print_endline msg
				| End_of_file ->
					Lex.print_position lex#demand_position;
					print_endline "unexpected end of text"
				| Not_found ->
					Lex.print_position lex#demand_position;
					print_endline "unkown identifier"
			done
		end 
end

module Main = Loop(Datas.ER_DiGraph(Datas.TaxonomyTree))
